#!/bin/bash
set -e

# Choose server id
[  -n "$1" ] && SID=$1 || SID=01

# Set parameters
NUM_INSTALLCHECKS=30
NUM_ITERATIONS=10

PGPREFIX="$(pwd)/test_install"

STORAGE=$(pwd)
SRCCOPYDIR="$STORAGE/icx"

PGPORT=5432; export PGPORT

[ "$SID" = "01" ] && PGDB="$(pwd)/tmpdb" || PGDB="$(pwd)/tmpdb$SID"
export PGDB

STMP=$STORAGE/s$SID/
[ -d "$STMP" ] && rm -rf "$STMP"
mkdir "$STMP"

if [ ! -d "$PGPREFIX" ]; then
  make -s distclean 2>/dev/null || true

  echo "Building postgres..."
  CPPFLAGS="$CPPFLAGS -Og -Wall -Wno-unused-parameter -Wno-sign-compare" \
  ./configure -q --prefix="$PGPREFIX" --enable-debug --enable-cassert --with-liburing
  make -s -j"$(nproc)" world-bin && make -s -j"$(nproc)" -C src/test && make -s install-world-bin && make -s install -C src/test && make -s install -C src/test/modules || exit 1
  echo "Done."
fi

export PGBIN="$PGPREFIX/bin"

export PATH="$PGBIN:$PATH"
export LD_LIBRARY_PATH="$PGPREFIX/lib"

export PGUSER=postgres

[ -d "$PGDB" ] || mkdir "$PGDB" 

# Stop previous instance in case it is left from a previous run
pg_ctl -w -t 300 -D "$PGDB" stop -m immediate || true

# And clear data and logs
rm -rf "$PGDB" "$STORAGE/server$SID.log" "installcheck$SID-"*.log || true

initdb -k -D "$PGDB" --username=postgres >initdb$SID.log 2>&1
echo "
port = $PGPORT
lc_messages = 'C'
fsync = off
autovacuum = off
restart_after_crash = off

io_max_concurrency = 2

io_method=io_uring

checkpoint_timeout = 30s
" >> "$PGDB/postgresql.auto.conf"

# Prepare template directory
[ -d "$SRCCOPYDIR" ] && rm -rf "$SRCCOPYDIR"
mkdir "$SRCCOPYDIR"
mkdir "$SRCCOPYDIR/template"
cp -p ./* "$SRCCOPYDIR/template" >/dev/null 2>&1 || true
echo "exit 0" > "$SRCCOPYDIR/template/config.status"
mkdir "$SRCCOPYDIR/template/src"
cp -p src/* "$SRCCOPYDIR/template/src/" >/dev/null 2>&1 || true
cp -p -r src/test "$SRCCOPYDIR/template/src/"
cp -p -r src/port "$SRCCOPYDIR/template/src/"
cp -p -r src/makefiles "$SRCCOPYDIR/template/src"

find . -name Makefile -print0 -o -name GNUmakefile -print0 | while IFS= read -r -d '' m; do
  mkdir -p "$SRCCOPYDIR/template/$(dirname "$m")"
  cp -p "$m" "$SRCCOPYDIR/template/$m"
done

sed -e 's/^all:/#all:/' -i "$SRCCOPYDIR/template/src/test/regress/GNUmakefile"

find contrib \( -name "*.conf" -o -name "*schedule" \) -exec cp -p "{}" "$SRCCOPYDIR/template/{}" \;

# Avoid rebuilding
sed -e "s/all: all-lib.*\$/all:/" -i "$SRCCOPYDIR/template/src/interfaces/libpq/Makefile"
sed -e "s/all:.*\$/all:/" -i "$SRCCOPYDIR/template/src/port/Makefile"
sed -e "s/all:.*\$/all:/" -i "$SRCCOPYDIR/template/src/common/Makefile"

find . -type d -name sql -print0 -o -name specs -print0 | while IFS= read -r -d '' d; do
  mkdir -p "$SRCCOPYDIR/template/$(dirname "$d")"
  cp -r "$d" "$SRCCOPYDIR/template/$(dirname "$d")/"
  [ -d "$d/../data" ] && cp -r "$d/../data" "$SRCCOPYDIR/template/$d/../"
  [ -d "$d/../expected" ] && cp -r "$d/../expected" "$SRCCOPYDIR/template/$d/../"
  [ -d "$d/../output" ] && cp -r "$d/../output" "$SRCCOPYDIR/template/$d/../"
  true
done

# Prepare separate installcheck directories
rm -rf "$SRCCOPYDIR/ic$SID"-* || true
for ((c=1;c<=NUM_INSTALLCHECKS;c++)); do
  mkdir "$SRCCOPYDIR/ic$SID-$c/"
  cp -p -r "$SRCCOPYDIR"/template/* "$SRCCOPYDIR/ic$SID-$c/"

  # Fix tests to use separate entities in different databases
  c03="$(printf "%03d" "$c")"

  sed -e "s/regression/regress$c03/" -i "$SRCCOPYDIR/ic$SID-$c/src/test/regress/expected/updatable_views.out"
  sed -e "s/regression/regress$c03/" -i "$SRCCOPYDIR/ic$SID-$c/src/test/regress/expected/sequence.out"
  sed -e "s/regression/regress$c03/" -i "$SRCCOPYDIR/ic$SID-$c/src/test/regress/expected/xmlmap.out"
  sed -e "s/regression/regress$c03/g" -i "$SRCCOPYDIR/ic$SID-$c/src/test/regress/expected/domain.out"
  sed -e "s/regression\./regress$c03./" -i "$SRCCOPYDIR/ic$SID-$c/src/test/regress/sql/psql.sql" \
    "$SRCCOPYDIR/ic$SID-$c/src/test/regress/expected/psql.out"

  sed -e "s/regression /regress$c03 /" -i "$SRCCOPYDIR/ic$SID-$c/src/test/regress/sql/privileges.sql" \
    "$SRCCOPYDIR/ic$SID-$c/src/test/regress/expected/privileges.out"

  sed -e "s/regression_/regress${c03}_/g" -i "$SRCCOPYDIR/ic$SID-$c/src/test/regress/sql/database.sql" \
    "$SRCCOPYDIR/ic$SID-$c/src/test/regress/expected/database.out"

  sed -e "s/ regression/ regress$c03/" -i "$SRCCOPYDIR/ic$SID-$c/src/test/regress/sql/publication.sql" \
    "$SRCCOPYDIR/ic$SID-$c/src/test/regress/expected/publication.out"

  sed -e "s/ regression/ regress$c03/i" -i "$SRCCOPYDIR/ic$SID-$c/src/test/regress/sql/subscription.sql" \
    "$SRCCOPYDIR/ic$SID-$c/src/test/regress/expected/subscription.out"

  sed -e "s/ regression/ regress$c03/" -i "$SRCCOPYDIR/ic$SID-$c/src/test/regress/sql/create_role.sql" \
    "$SRCCOPYDIR/ic$SID-$c/src/test/regress/expected/create_role.out"

  sed -e "s/ regression/ regress$c03/" -i "$SRCCOPYDIR/ic$SID-$c/src/test/regress/sql/dependency.sql" \
    "$SRCCOPYDIR/ic$SID-$c/src/test/regress/expected/dependency.out"

  # Adjust regress entities
  find "$SRCCOPYDIR/ic$SID-$c/" \( -name "*.sql" -o -name "*.source" -o -name "*.out" -o -name "*.spec" \) -exec sed -e " \
s/regress_\(tblspace\)/regr${c03}_\1/g; \
s/regress_\(vacuum\)/regr${c03}_\1/g; \
s/regress_\(testsub\)/regr${c03}_\1/g; \
s/regress_\(regrole_test\)/regr${c03}_\1/g; \
s/regress_\(stats_ext\)/regr${c03}_\1/g; \
s/regress_\(constraint_comments\)/regr${c03}_\1/g; \
s/regress_\(no_child_access\)/regr${c03}_\1/g; \
s/regress_\(rol_lock1\)/regr${c03}_\1/g; \
s/regress_\(nosuper\|view_owner\)/regr${c03}_\1/g; \
s/regress_\(truncate_conflict\)/regr${c03}_\1/g; \
s/regress_\(alice\|bob\|carol\|dave\)/regr${c03}_\1/g; \
s/regress_\(merge_privs\|merge_no_privs\|merge_none\)/regr${c03}_\1/g; \
s/regress_\(maintain\|no_maintain\|grantor\)/regr${c03}_\1/g; \
s/regress_\(readallstats\)/regr${c03}_\1/g; \
s/regress_\(createdb\|createrole\|login\|inherit\|connection_limit\|encrypted_password\|password_null\|noiseword\|inroles\|adminroles\|rolecreator\)/regr${c03}_\1/g; \
s/regress_\(read_all_data\|write_all_data\|monitor\|read_all_settings\|read_all_stats\|stat_scan_tables\|read_server_files\|write_server_files\|execute_server_program\|signal_backend\)/regr${c03}_\1/g; \
s/regress_\(tenant\|hasprivs\|role_admin\|role_super\)/regr${c03}_\1/g; \
s/regress_\( \)/regr${c03}_\1/g; \
s/regress_\(ptnowner\)/regr${c03}_\1/g; \
s/regress_\(create_idx_tblspace\)/regr${c03}_\1/g; \
s/regress_\(log_memory\|slot_dir_funcs\|current_logfile\)/regr${c03}_\1/g; \
s/regress_\(bttest_role\)/regr${c03}_\1/g; \
s/regress_\(cfs_tblspace\)/regr${c03}_\1/g; \
s/regress_\([a-z_]*user[0-9]*\)/regr${c03}_\1/g; \
s/regress_\([a-z_]*role\)/regr${c03}_\1/g; \
s/regress_\([a-z_]*group\)/regr${c03}_\1/g; \
s/regress_\([a-z_]*admin\)/regr${c03}_\1/g; \
s/regress_\([a-z_]*owner\)/regr${c03}_\1/g; \
s/regress_\(test_[a-z0-9_]*\)/regr${c03}_\1/g; \
s/regress_\(priv_[a-z0-9]\+\)/regr${c03}_\1/g; \
s/regress_\(rls_[a-z0-9]\+\)/regr${c03}_\1/g; \
s/regress_\(r1\|s1\)/regr${c03}_\1/g; \
s/regress_\(admin\|host_resource_admin\)/regr${c03}_\1/g; \
s/regress_\(cluster_conflict\|cluster_part\)/regr${c03}_\1/g; \
s/regress_\(bypassrls\|replication\|replication_bypassrls\)/regr${c03}_\1/g; \
s/regress_\(datdba_before\|datdba_after\)/regr${c03}_\1/g; \
s/regress_\(groot\|outis\)/regr${c03}_\1/g; \
s/isolation_\(slot\)/isolat${c03}_\1/g; \
s/user_\(cat\)/u${c03}_\1/g; \
s/regress_\(passwd[a-z0-9_]*\)/regr${c03}_\1/g; " -i {} \;
done

ulimit -c unlimited
ulimit -n 4096

# Start server
pg_ctl -w -t 15 -D "$PGDB" -l "$STORAGE/server$SID.log" start

# Run `make installcheck`s in a loop
result=0
time for ((i=1;i<=NUM_ITERATIONS;i++)); do

  echo "iteration $i: $(date)"
  for ((c=1;c<=NUM_INSTALLCHECKS;c++)); do
    c03=$(printf "%03d" "$c")
    rm -rf "$SRCCOPYDIR/ic$SID-$c/src/test/regress/results"/*
    (cd "$SRCCOPYDIR/ic$SID-$c"; EXTRA_REGRESS_OPTS="--dbname=regress$c03" EXTRA_TESTS="--max-connections=5" make -s installcheck NO_GENERATED_HEADERS=1 USE_MODULE_DB=1 || echo FAIL) >>"installcheck$SID-$c.log" 2>&1 &
  done
  wait
  echo "installchecks finished: $(date)"

  if grep 'could not read block' "$STORAGE/server$SID.log"; then
    echo "FAILED"
    result=1
    break
  fi

done

pg_ctl -w -t 90 -D "$PGDB" stop
exit $result
